#' command-line interface to \code{\link{jlim.test}}
#'
#' See accompanied run_jlim.sh to execute this function. 
#'
#' Usage:
#' \code{run_jlim.sh input.cfg.tsv r2_resolution outfile}
#'
#' Command line arguments:
#' \itemize{
#' \item \code{input.cfg.tsv}: config file generated by \code{\link{jlim.gencfg}}
#' \item \code{r2_resolution}: r2 resolution between 0.0 and 1.0 (default 0.8)
#' \item \code{outfile}: output file. Table with two column STAT and p. STAT is JLIM statistic lambda. p is the permutation p-value.
#' }
#' @export
run.jlim <- function() {

    argvals <- commandArgs(trailingOnly=TRUE)

    if (length(argvals) != 3) {
        stop("Usage: run_jlim.sh cfg.tsv r2_resolution output.tsv")
    }

    cfgfile <- argvals[1]
    r2res <- as.numeric(argvals[2])
    outfile <- argvals[3]

    if (is.na(r2res) || r2res <= 0 || r2res >= 1) {
        stop(paste("invalid r2 resolution (0-1.0):", argvals[2]))
    }

    if (!file.exists(cfgfile)) {
        cat(paste("File does not exist:", cfgfile, "\n"))
        q(status=1)
    }

    cfg <- read.delim(cfgfile, sep="\t", stringsAsFactors=FALSE, header=TRUE)

    if (!("idxSNP" %in% colnames(cfg))) {
        # missing header
        # reload
        cfg <- read.delim(cfgfile, sep="\t", stringsAsFactors=FALSE,
                          header=FALSE)

        # default cfg columns
#        cfgcol <- c("tr1name", "chrom", "idxSNP", "idxBP", "idxP",
#                    "idx2BP", "idx2P", "start", "end",
#                    "tr2name", "tr2subID",
#                    "minP2", "maintr", "sectr", "refld", "mainld", "secld",
#                    "perm")
        cfgcols <- c("maintrID", "chrom", "idxSNP",  "idxBP", "idxP", 
                     "idx2BP", "idx2P", "start", "end",
                     "sectrID", "sectrsubID",
                     "minP2", "maintr", "sectr", "refld", "mainld", "secld",
                     "perm")
                      
        if(ncol(cfg) != length(cfgcols)) {
            stop(paste("Unrecognizable config file:", cfgfile))
        }

        colnames(cfg) <- cfgcols
    }

    df.STAT <- c()
    df.p <- c()

    for (I in 1:nrow(cfg)) {
        PL("main trait ID", cfg$maintrID[I])
        PL("index SNP ID", cfg$idxSNP[I])
        PL("secondary trait ID", cfg$sectrID[I])
        PL("secondary trait subID", cfg$sectrsubID[I])
        PL("main trait summary stat file", cfg$maintr[I])
        PL("secondary trait summary stat file", cfg$sectr[I])
        PL("reference LD file", cfg$refld[I])
        if (cfg$mainld[I] != ".") {
            PL("main trait LD file", cfg$mainld[I])
        }
        PL("secondary trait LD file", cfg$secld[I])
        PL("secondary trait permutation file", cfg$perm[I])
        PL("start (BP)", cfg$start[I])
        PL("end (BP)", cfg$end[I])
        PL("r2 resolution", r2res)
        
        jlimout <- 
            jlim.test(cfg$maintr[I], cfg$sectr[I], cfg$refld[I],
                      cfg$secld[I], cfg$perm[I], cfg$start[I], cfg$end[I],
                      r2res, cfg$mainld[I])

        df.STAT <- c(df.STAT, jlimout@STAT)
        df.p <- c(df.p, jlimout@p.value)
    }

    write.table(data.frame(STAT=df.STAT, p=df.p),
                file=outfile, quote=FALSE, sep="\t", row.names=FALSE,
                col.names=TRUE)
        
}


